"""
Jabber connection class
Basicly only shuffling messages to and from the network but
also manages some tasks like version replies and roster management.
"""
import sys
import time
import socket
import random
import types

import jabber.jabber as jabber
import BApplication
import BMessage
import BMessageRunner
import BMessenger
import BLooper

from AppKit import *

from modules.log import log
from modules.constants import *
from modules.util import *

import canned
import callback

NOT_CONNECTED_DESCR = 'Not connected.'
###########
### Classes
###########

class Worker(BLooper.BLooper):
	def __init__(self):
		BLooper.BLooper.__init__(self, 'Connection Worker')
		self.connected = False
		self.Run()

	def QuitRequested(self):
		self.disconnect()
		log('Connection Worker quit')
		return True

	def MessageReceived(self, msg):
		if msg.what == PROCESS_CONNECTION:
			if self.connected:
				try:
					self.client.process(0)
				except socket.error, (errno, errstr):
					log('socket error: %s' % errstr)
					BApplication.be_app.PostMessage(ErrorMessage(errstr))
					self.disconnect()
		elif msg.what == OUTBOUND_PRESENCE:
			if self.connected:
				try:
					try:
						show = msg.FindString('show')
					except BMessage.error:
						show = None
					try:
						status = msg.FindString('status')
					except BMessage.error:
						status = None
					try:
						ptype = msg.FindString('type')
					except BMessage.error:
						ptype = None
					try:
						pri = msg.FindString('priority')
					except BMessage.error:
						pri = None
					try:
						# if there's a to field we need to send manually
						pto = msg.FindString('to')
					except BMessage.error:
						self.client.sendPresence(ptype,pri,show,status)
					else:
						prs = jabber.Presence(to=pto, type=ptype)
						self.client.send(prs)
				except socket.error, (errno, errstr):
					log('socket error: %s' % errstr)
					msg.SendReply(ErrorMessage(errstr))
					self.disconnect()
			else:
				msg.SendReply(ErrorMessage(NOT_CONNECTED_DESCR))
		elif msg.what == OUTBOUND_MESSAGE:
			if not self.connected:
				if msg.IsSourceWaiting():
					msg.SendReply(ErrorMessage('Not connected.'))
				return
			try:
				mtype = msg.FindString('type')
			except BMessage.error:
				mtype = 'normal'
			if mtype != 'chat':
				jmsg = jabber.Message(msg.FindString('to'), msg.FindString('body'), 'normal', msg.FindString('subject'))
			else:
				jmsg = jabber.Message(msg.FindString('to'), msg.FindString('body'), 'chat',)
			try:
				self.client.send(jmsg)
			except socket.error, (errno, errstr):
				log('socket error: %s' % errstr)
				if msg.IsSourceWaiting():
					msg.SendReply(ErrorMessage(errstr))
				self.disconnect()
			if msg.IsSourceWaiting():
				msg.SendReply(BMessage.BMessage(STATUS_OK))
		elif msg.what == MSG_CONNECT:
			self.connect(msg)
		elif msg.what == MSG_DISCONNECT:
			self.disconnect()
		elif msg.what == ROSTER_ITEM_REQUEST:
			if not self.connected:
				if msg.IsSourceWaiting():
					msg.SendReply(ErrorMessage('Not connected.'))
				return
			jid = msg.FindString('jid')
			roster = self.client.getRoster()
			if jid in roster.getJIDs():
				bmsg = BMessage.BMessage(STATUS_OK)
				online = roster.getOnline(jid)
				bmsg.AddString('online', toUTF8(online))
				if online == 'online':
					show = roster.getShow(jid)
					if show is not None:
						bmsg.AddString('show', toUTF8(show))
					else:
						bmsg.AddString('show', 'available')
					status = roster.getStatus(jid)
					if status is not None:
						bmsg.AddString('status', toUTF8(status))
				sub = roster.getSub(jid)
				if sub is not None:
					bmsg.AddString('sub', sub)
				msg.SendReply(bmsg)
			else:
				msg.SendReply(ErrorMessage('Not in roster.'))

	####
	def connect(self, msg):
		if not self.connected:
			host = msg.FindString('host')
			port = msg.FindInt32('port')
			self.isSSL =  False
			if msg.FindBool('ssl'):
				port = 5223
				self.client = jabber.Client(host=host, port=port, connection=2)
				self.isSSL = True
			else:
				self.client = jabber.Client(host=host, port=port, connection=1)
			self.client.setDisconnectHandler(self._disconnectHandler)
			user = msg.FindString('user')
			password = msg.FindString('passwd')
			resource = msg.FindString('resource')
			try:
				self.client.connect()
			except socket.error, (errno, errstr):
				log('socket error: %s' % errstr)
				msg.SendReply(ErrorMessage(errstr))
				return
			if not self.client.auth(user, password, resource):
				try:
					self.disconnect()
				except socket.error, (errno, errstr):
					log('socket error: %s' % errstr)
				msg.SendReply(ErrorMessage('Wrong username or password'))
				return
			self.connected = True
			# Register message handler
			self.client.registerHandler('message',callback.Message)
			self.client.registerHandler('iq',callback.Iq)
			self.client.registerHandler('presence',callback.Presence)
			# Request roster
			roster = self.client.requestRoster()
			roster.setListener(callback.RosterListener)
			# Set up the interval for checking new msgs from server.
			# Use a messenger that sends a message every X (pico?)seconds
			# so that we can use beos messaging and dont have to use
			# python threads and queues.
			message = BMessage.BMessage(PROCESS_CONNECTION)
			messenger = BMessenger.BMessenger(self)
			self.runner = BMessageRunner.BMessageRunner(messenger, message, 500000, -1)
			self.runner.InitCheck()
			# Send initial presence
			try:
				self.client.sendInitPresence()
			except socket.error, (errno, errstr):
				log('socket error: %s' % errstr)
				BApplication.be_app.PostMessage(ErrorMessage(errstr))
				self.disconnect()
				return
			# Signal connect back to main
			rmsg = BMessage.BMessage(STATUS_CONNECTED)
			rmsg.AddString('jid', '%s@%s' % (user,host))
			rmsg.AddString('resource', '%s' % (resource))
			rmsg.AddBool('ssl', self.isSSL)
			BApplication.be_app.PostMessage(BMessage.BMessage(rmsg))

	def disconnect(self):
		if self.connected:
			# Need to handle SSL sockets manually since
			# jabberpy doesnt handle them correctly.
			if not self.isSSL:
				try:
					self.client.disconnect()
				except socket.error, (errno, errstr):
					log('socket error: %s' % errstr)
			else:
				self.client.send(jabber.Presence(type='unavailable'))
				sock = self.client.getSocket()
				sock.send(u"</stream:stream>")
				try:
					sock.shutdown(2)
				except socket.error:
					log('error doing shutdown() on socket')
				sock.close()
				self._disconnectHandler(self.client)
				
			self.connected = False

	def _disconnectHandler(self, con):
		self.connected = False
		bmsg = BMessage.BMessage(STATUS_DISCONNECTED)
		# does lastErr work?
		#print con.lastErr
		#bmsg.AddString('errstr', con.lastErr())
		BApplication.be_app.PostMessage(bmsg)
